"""
.. currentmodule:: neet.boolean.logicnetwork

.. testsetup:: logicnetwork

    from neet.boolean.logicnetwork import LogicNetwork

Logic-based Networks
====================
"""
import re
from neet.python import long
from neet.statespace import StateSpace
from neet.exceptions import FormatError
import networkx as nx


class LogicNetwork(object):
    """
    The LogicNetwork class represents boolean networks whose update rules
    follow logic relations among nodes. Each node state is expressed as ``0``
    or ``1``.
    """

    def __init__(self, table, names=None, reduced=False):
        """
        Construct a network from a logic truth table.

        A truth table stores a list of tuples, one for each node in order. A
        tuple of the form `(A, {C1, C2, ...})` at index `i` provides the
        activation conditions for the node of index `i`. `A` is a tuple marking
        the indices of the nodes which influence the state of node `i` via
        logic relations. `{C1, C2, ...}` is a set, each element of which is the
        collection of binary states of these influencing nodes that would
        activate node `i`, setting it to `1`. Any other collection of states of
        nodes in `A` are assumed to deactivate node `i`, setting it to `0`.

        `C1`, `C2`, etc. are sequences (`tuple` or `str`) of binary digits,
        each being the binary state of corresponding node in `A`.

        .. rubric:: Examples

        .. doctest:: logicnetwork

            >>> net = LogicNetwork([((0,), {'0'})])
            >>> net.size
            1
            >>> net.table
            [((0,), {'0'})]

        .. doctest:: logicnetwork

            >>> net = LogicNetwork([((1,), {'0', '1'}), ((0,), {'1'})])
            >>> net.size
            2
            >>> net.table == [((1,), {'0', '1'}), ((0,), {'1'})]
            True

        .. doctest:: logicnetwork

            >>> net = LogicNetwork([((1, 2), {'01', '10'}),
            ... ((0, 2), ((0, 1), '10', [1, 1])),
            ... ((0, 1), {'11'})], ['A', 'B', 'C'])
            >>> net.size
            3
            >>> net.names
            ['A', 'B', 'C']
            >>> net.table == [((1, 2), {'01', '10'}),
            ... ((0, 2), {'01', '11', '10'}), ((0, 1), {'11'})]
            True

        :param table: the logic table
        :param names: names of nodes, default None
        """
        if not isinstance(table, (list, tuple)):
            raise TypeError("table must be a list or tuple")

        self.__size = len(table)

        if names:
            if not isinstance(names, (list, tuple)):
                raise TypeError("names must be a list or tuple")
            elif len(names) != self.__size:
                raise ValueError("number of names must match network size")
            else:
                self.names = list(names)

        # Store positive truth table for human reader.
        self.table = []
        for row in table:
            # Validate incoming indices.
            if not (isinstance(row, (list, tuple)) and len(row) == 2):
                raise ValueError("Invalid table format")
            for idx in row[0]:
                if idx >= self.__size:
                    raise IndexError("mask index out of range")
            # Validate truth table of the sub net.
            if not isinstance(row[1], (list, tuple, set)):
                raise ValueError("Invalid table format")
            conditions = set()
            for condition in row[1]:
                conditions.add(''.join([str(long(s)) for s in condition]))
            self.table.append((row[0], conditions))

        if reduced:
            self.reduce_table()

        self._state_space = StateSpace(self.__size, base=2)

        # Encode truth table for faster computation.
        self._encode_table()

        self.metadata = {}

    @property
    def size(self):
        """
        The number of nodes in the network.

        .. doctest:: logicnetwork

            >>> net = LogicNetwork([((1, 2), {'01', '10'}),
            ... ((0, 2), {'01', '10', '11'}), ((0, 1), {'11'})])
            >>> net.size
            3

        :type: int
        """
        return self.__size

    def _encode_table(self):
        self._encoded_table = []
        for indices, conditions in self.table:
            # Encode the mask.
            mask_code = long(0)
            for idx in indices:
                mask_code += 2 ** long(idx)  # Low order, low index.
            # Encode each condition of truth table.
            encoded_sub_table = set()
            for condition in conditions:
                encoded_condition = long(0)
                for idx, state in zip(indices, condition):
                    encoded_condition += 2 ** long(idx) if int(state) else 0
                encoded_sub_table.add(encoded_condition)
            self._encoded_table.append((mask_code, encoded_sub_table))

    def is_dependent(self, target, source):
        """
        Return True if state of ``target`` is influenced by the state of
        ``source``.

        .. doctest:: logicnetwork

            >>> net = LogicNetwork([((1, 2), {'01', '10'}),
            ... ((0, 2), {'01', '10', '11'}), ((0, 1), {'11'})])
            >>> net.is_dependent(0, 0)
            False
            >>> net.is_dependent(0, 2)
            True

        :param target: index of the target node
        :param source: index of the source node
        :returns: whether the target node is dependent on the source
        """
        sub_table = self.table[target]
        if source not in sub_table[0]:  # No explicit dependency.
            return False

        # Determine implicit dependency.
        i = sub_table[0].index(source)
        counter = {}
        for state in sub_table[1]:
            # State excluding source.
            state_sans_source = state[:i] + state[i + 1:]
            if long(state[i]) == 1:
                counter[state_sans_source] = counter.get(
                    state_sans_source, 0) + 1
            else:
                counter[state_sans_source] = counter.get(
                    state_sans_source, 0) - 1

        if any(counter.values()):  # States uneven.
            return True
        return False

    def reduce_table(self):
        """
        Reduce truth table by removing input nodes which have no logic
        influence from the truth table of each node.

        .. note::
            This function introduces the identity function for all nodes which
            have no inputs. This ensure that every node has a well-defined
            logical function. The example below demonstrates this with node
            ``1``.

        .. doctest:: logicnetwork

            >>> net = LogicNetwork([((0,1), {'00', '10'}), ((0,), {'0', '1'})])
            >>> net.table == [((0,1), {'00', '10'}), ((0,), {'0', '1'})]
            True
            >>> net.reduce_table()
            >>> net.table == [((1,), {'0'}), ((1,), {'1'})]
            True
        """
        reduced_table = []
        for node, (sources, conditions) in enumerate(self.table):
            reduced_sources = []
            reduced_indices = []
            for idx, source in enumerate(sources):
                if self.is_dependent(node, source):
                    reduced_sources.append(source)
                    reduced_indices.append(idx)

            if reduced_sources:  # Node state is influenced by other nodes.
                reduced_conditions = set()
                for condition in conditions:
                    reduced_condition = ''.join([str(condition[idx])
                                                 for idx in reduced_indices])
                    reduced_conditions.add(reduced_condition)
            else:
                # Node state is not influenced by other nodes including itself.
                reduced_sources = (node, )
                if not conditions:
                    # If original conditions is empty, node is never activated.
                    reduced_conditions = set()
                elif node in sources:
                    # Node is always activated no matter its previous state.
                    reduced_conditions = {'0', '1'}
                else:
                    # Node state is not changed.
                    reduced_conditions = {'1'}

            reduced_table.append((tuple(reduced_sources), reduced_conditions))

        self.table = reduced_table

        self._encode_table()

    def state_space(self):
        """
        Return a :class:`neet.statespace.StateSpace` object for the network.

        .. doctest:: logicnetwork

            >>> net = LogicNetwork([((1, 2), {'01', '10'}),
            ... ((0, 2), {'01', '10', '11'}), ((0, 1), {'11'})])
            >>> net.state_space()
            <neet.statespace.StateSpace object at 0x...>
            >>> space = net.state_space()
            >>> list(space)
            [[0, 0, 0], [1, 0, 0], [0, 1, 0], [1, 1, 0], [0, 0, 1], [1, 0, 1], [0, 1, 1], [1, 1, 1]]

        :returns: the network's :class:`neet.statespace.StateSpace`
        """
        return self._state_space

    def _unsafe_update(self, net_state, index=None, pin=None, values=None):
        """
        Unsafely update node states according to the truth table.

        If ``index`` is provided, only update the node at ``index``. If
        ``index`` is not provided, update all nodes. The input ``net_state`` is
        not modified.

        .. rubric:: Examples

        .. doctest:: logicnetwork

            >>> net = LogicNetwork([((0,), {'0'})])
            >>> net._unsafe_update([0], 0)
            [1]
            >>> net._unsafe_update([1])
            [0]

        .. doctest:: logicnetwork

            >>> net = LogicNetwork([((1,), {'0', '1'}), ((0,), {'1'})])
            >>> net._unsafe_update([1, 0], 0))
            [1, 0]
            >>> net._unsafe_update([1, 0], 1))
            [1, 1]
            >>> net._unsafe_update([0, 0])
            [1, 0]

        .. doctest:: logicnetwork

            >>> net = LogicNetwork([((1, 2), {'01', '10'}),
            ... ((0, 2), {(0, 1), '10', (1, 1)}),
            ... ((0, 1), {'11'})])
            >>> net.size
            3
            >>> net._unsafe_update([0, 1, 0])
            [1, 0, 0]
            >>> net._unsafe_update([0, 0, 1])
            [1, 1, 0]
            >>> net._unsafe_update([0, 0, 1], 1)
            [0, 1, 1]
            >>> net._unsafe_update([0, 0, 1], pin=[1])
            [1, 0, 0]
            >>> net._unsafe_update([0, 0, 1], pin=[0, 1])
            [0, 0, 0]
            >>> net._unsafe_update([0, 0, 1], values={0: 0})
            [0, 1, 0]
            >>> net._unsafe_update([0, 0, 1], pin=[1], values={0: 0})
            [0, 0, 0]

        :param net_state: a sequence of binary node states
        :type net_state: sequence
        :param index: the index to update (or None)
        :type index: int or None
        :param pin: the indices to pin (or None)
        :type pin: sequence
        :param values: override values
        :type values: dict
        :returns: the updated states
        """
        encoded_state = self.state_space()._unsafe_encode(net_state)

        if index is None:
            indices = range(self.__size)
        else:
            indices = [index]

        if pin is None:
            pin = []

        for idx in indices:
            if idx in pin:
                continue
            mask, condition = self._encoded_table[idx]
            sub_net_state = mask & encoded_state
            net_state[idx] = 1 if sub_net_state in condition else 0

        if values:
            for k, v in values.items():
                net_state[k] = v

        return net_state

    def update(self, net_state, index=None, pin=None, values=None):
        """
        Update node states according to the truth table.

        If ``index`` is provided, only update the node at ``index``. If
        ``index`` is not provided, update all ndoes. ``pin`` provides the
        indices of which the nodes' states are forced to remain unchanged.

        .. rubric:: Examples

        .. doctest:: logicnetwork

            >>> net = LogicNetwork([((0,), {'0'})])
            >>> net.update([0], 0)
            [1]
            >>> net.update([1])
            [0]
            >>>

        .. doctest:: logicnetwork

            >>> net = LogicNetwork([((1,), {'0', '1'}), ((0,), {'1'})])
            >>> net.update([1, 0], 0)
            [1, 0]
            >>> net.update([1, 0], 1)
            [1, 1]
            >>> net.update([0, 0])
            [1, 0]

        .. doctest:: logicnetwork

            >>> net = LogicNetwork([((1, 2), {'01', '10'}),
            ... ((0, 2), {(0, 1), '10', (1, 1)}),
            ... ((0, 1), {'11'})])
            >>> net.size
            3
            >>> net.update([0, 0, 1], 1)
            [0, 1, 1]
            >>> net.update([0, 1, 0])
            [1, 0, 0]
            >>> net.update([0, 0, 1])
            [1, 1, 0]
            >>> net.update([0, 0, 1], pin=[1])
            [1, 0, 0]
            >>> net.update([0, 0, 1], pin=[0, 1])
            [0, 0, 0]
            >>> net.update([0, 0, 1], values={0: 0})
            [0, 1, 0]
            >>> net.update([0, 0, 1], pin=[1], values={0: 0})
            [0, 0, 0]

        :param net_state: a sequence of binary node states
        :type net_state: sequence
        :param index: the index to update (or None)
        :type index: int or None
        :param pin: the indices to pin (or None)
        :type pin: sequence
        :param values: override values
        :type values: dict
        :returns: the updated states
        """
        if net_state not in self.state_space():
            raise ValueError(
                "the provided state is not in the network's state space")

        if values and any([v not in (0, 1) for v in values.values()]):
            raise ValueError("invalid state in values argument")

        if pin and values and any([k in pin for k in values]):
            raise ValueError("cannot set a value for a pinned state")

        return self._unsafe_update(net_state, index, pin, values)

    @classmethod
    def read_table(cls, table_path, reduced=False):
        """
        Read a network from a truth table file.

        A logic table file starts with a table title which contains names of
        all nodes. It is a line marked by ``##`` at the begining with node
        names seperated by commas or spaces. This line is required. For
        artificial network without node names, arbitrary names must be put in
        place, e.g.:

        ``## A B C D``

        Following are the sub-tables of logic conditions for every node. Each
        sub-table nominates a node and its logically connected nodes in par-
        enthesis as a comment line:

        ``# A (B C)``

        The rest of the sub-table are states of those nodes in parenthesis
        ``(B, C)`` that would activate the state of A. States that would
        deactivate ``A`` should not be included in the sub-table.

        A complete logic table with 3 nodes A, B, C would look like this:

        ::

            ## A B C
            # A (B C)
            1 0
            1 1
            # B (A)
            1
            # C (B C A)
            1 0 1
            0 1 0
            0 1 1

        Custom comments can be added above or below the table title (as long as
        they are preceeded with more or less than two ``#`` (eg ``#`` or
        ``###`` but not ``##``)).

        .. rubric:: Examples:

        .. doctest:: logicnetwork

            >>> myeloid_path = '../neet/boolean/data/myeloid-truth_table.txt'
            >>> net = LogicNetwork.read_table(myeloid_path)
            >>> net.size
            11
            >>> net.names
            ['GATA-2', 'GATA-1', 'FOG-1', 'EKLF', 'Fli-1', 'SCL', 'C/EBPa', 'PU.1', 'cJun', 'EgrNab', 'Gfi-1']

        :param table_path: a path to a table table file
        :type table_path: str
        :returns: a :class:`LogicNetwork`
        """
        names_format = re.compile(r'^\s*##[^#]+$')
        node_title_format = re.compile(
            r'^\s*#\s*(\S+)\s*\((\s*(\S+\s*)+)\)\s*$')

        with open(table_path, 'r') as f:
            lines = f.read().splitlines()
            # Search for node names.
            i = 0
            names = []
            while not names:
                try:
                    if names_format.match(lines[i]):
                        names = re.split(r'\s*,\s*|\s+', lines[i].strip())[1:]
                    i += 1
                except IndexError:
                    raise FormatError("node names not found in file")

            table = [()] * len(names)
            # Create condition tables for each node.
            for line in lines[i:]:
                node_title = node_title_format.match(line)
                if node_title:
                    node_name = node_title.group(1)
                    # Read specifications for node.
                    if node_name not in names:
                        raise FormatError(
                            "'{}' not in node names".format(node_name))
                    node_index = names.index(node_name)
                    sub_net_nodes = re.split(
                        r'\s*,\s*|\s+', node_title.group(2).strip())

                    in_nodes = tuple(map(names.index, sub_net_nodes))
                    table[node_index] = (in_nodes, set())
                elif re.match(r'^\s*#.*$', line):
                    # Skip a comment.
                    continue
                else:
                    # Read activation conditions for node.
                    try:
                        if line.strip():
                            condition = re.split(r'\s*,\s*|\s+', line.strip())
                        else:
                            # Skip an empty line.
                            continue

                        if len(condition) != len(table[node_index][0]):
                            raise FormatError(
                                "number of states and nodes must match")
                        for state in condition:
                            if state not in ('0', '1'):
                                raise FormatError("node state must be binary")
                        table[node_index][1].add(''.join(condition))

                    except NameError:  # node_index not defined
                        raise FormatError(
                            "node must be specified before logic conditions")

        # If no truth table is provided for a node, that node is considered
        # an "external" node, i.e, its state stays on or off by itself.
        for i, sub_table in enumerate(table):
            if not sub_table:  # Empty truth table.
                table[i] = ((i,), {'1'})

        return cls(table, names, reduced)

    @classmethod
    def read_logic(cls, logic_path, external_nodes_path=None, reduced=False):
        """
        Read a network from a file of logic equations.

        A logic equations has the form of ``A = B AND ( C OR D )``, each term
        being separated from parantheses and logic operators with at least a
        space. The optional ``external_nodes_path`` takes a file that contains
        nodes in a column whose states do not depend on any nodes. These are
        considered "external" nodes. Equivalently, such a node would have a
        logic equation ``A = A``, for its state stays on or off unless being
        set externally, but now the node had to be excluded from
        ``external_nodes_path`` to avoid duplication and confusion.

        .. rubric:: Examples

        .. doctest:: logicnetwork

            >>> myeloid_path = '../neet/boolean/data/myeloid-logic_expressions.txt'
            >>> net = LogicNetwork.read_logic(myeloid_path)
            >>> net.size
            11
            >>> net.names
            ['GATA-2', 'GATA-1', 'FOG-1', 'EKLF', 'Fli-1', 'SCL', 'C/EBPa', 'PU.1', 'cJun', 'EgrNab', 'Gfi-1']

        :param logic_path: path to a file of logial expressions
        :type logic_path: str
        :param external_nodes_path: a path to a file of external nodes
        :type external_nodes_path: str
        :returns: a :class:`LogicNetwork`
        """
        names = []
        expressions = []
        with open(logic_path) as eq_file:
            for eq in eq_file:
                name, expr = eq.split('=')
                names.append(name.strip())
                expressions.append(expr.strip())

        if external_nodes_path:
            with open(external_nodes_path) as extra_file:
                extras = [name.strip() for name in extra_file]
            names += extras

        ops = {'AND', 'OR', 'NOT'}

        table = []
        for expr in expressions:
            sub_nodes = []
            conditions = set()

            expr_split = expr.split()
            for i, item in enumerate(expr_split):
                if item not in ops and item not in '()':
                    if item not in names:
                        raise ValueError("unknown component '{}'".format(item))
                    if item not in sub_nodes:
                        expr_split[i] = '{' + str(len(sub_nodes)) + '}'
                        sub_nodes.append(item)
                    else:
                        expr_split[i] = '{' + str(sub_nodes.index(item)) + '}'
                else:
                    expr_split[i] = item.lower()
            logic_expr = ' '.join(expr_split)

            indices = tuple([names.index(node) for node in sub_nodes])

            for dec_state in range(2**len(sub_nodes)):
                bin_state = '{0:0{1}b}'.format(dec_state, len(sub_nodes))
                if eval(logic_expr.format(*bin_state)):
                    conditions.add(bin_state)

            table.append((indices, conditions))

        # Add empty logic tables for external components.
        if external_nodes_path:
            for i in range(len(extras)):
                table.append((((len(names) - len(extras) + i),), set('1')))

        return cls(table, names, reduced)

    def neighbors_in(self, index):
        """
        Return the set of all neighbor nodes, where edge(neighbor_node-->index)
        exists.

        .. rubric:: Examples

        .. doctest:: logicnetwork

            >>> net = LogicNetwork([((1, 2), {'11', '10'}),
            ... ((0,), {'1'}),
            ... ((0, 1, 2), {'010', '011', '101'}),
            ... ((3,), {'1'})])
            >>> [net.neighbors_in(node) for node in range(net.size)]
            [{1, 2}, {0}, {0, 1, 2}, {3}]

        :param index: node index
        :returns: the set of all node indices which point toward the index node
        """
        return set(self.table[index][0])

    def neighbors_out(self, index):
        """
        Return the set of all neighbor nodes, where edge(index-->neighbor_node)
        exists.

        .. rubric:: Examples

        .. doctest:: logicnetwork


            >>> net = LogicNetwork([((1, 2), {'11', '10'}),
            ... ((0,), {'1'}),
            ... ((0, 1, 2), {'010', '011', '101'}),
            ... ((3,), {'1'})])
            >>> [net.neighbors_out(node) for node in range(net.size)]
            [{1, 2}, {0, 2}, {0, 2}, {3}]

        :param index: node index
        :returns: the set of all node indices which the index node points to
        """
        outgoing_neighbors = set()
        for i, incoming_neighbors in enumerate([row[0] for row in self.table]):
            if index in incoming_neighbors:
                outgoing_neighbors.add(i)

        return outgoing_neighbors

    def neighbors(self, index):
        """
        Return a set of neighbors for a specified node, or a list of sets of
        neighbors for all nodes in the network.

        .. rubric:: Examples

        .. doctest:: logicnetwork

            >>> net = LogicNetwork([((1, 2), {'11', '10'}),
            ... ((0,), {'1'}),
            ... ((0, 1, 2), {'010', '011', '101'}),
            ... ((3,), {'1'})])
            >>> [net.neighbors(node) for node in range(net.size)]
            [{1, 2}, {0, 2}, {0, 1, 2}, {3}]

        :param index: node index
        :returns: a set of neighbors of a node
        """
        return self.neighbors_in(index) | self.neighbors_out(index)

    def to_networkx_graph(self, labels='indices'):
        """
        Return networkx graph given neet network.

        :param labels: how node is labeled and thus identified in networkx
                       graph (``'names'`` or ``'indices'``)
        :returns: a ``networkx.DiGraph``
        """
        if labels == 'names':
            if hasattr(self, 'names') and (self.names is not None):
                labels = self.names
            else:
                raise ValueError("network nodes do not have names")

        elif labels == 'indices':
            labels = range(self.__size)

        else:
            raise ValueError("labels must be 'names' or 'indices'")

        edges = []
        for i, label in enumerate(labels):
            for j in self.neighbors_out(i):
                edges.append((labels[i], labels[j]))

        return nx.DiGraph(edges, name=self.metadata.get('name'))

    def draw(self, labels='indices', filename=None):
        """
        Output a file with a simple network drawing.

        Requires ``networkx`` and ``pygraphviz``.

        Supported image formats are determined by ``graphviz``. In particular,
        pdf support requires ``'cairo'`` and ``'pango'`` to be installed prior
        to graphviz installation.

        :param labels: how node is labeled and thus identified in networkx
                       graph (``'names'`` or ``'indices'``), only used if
                       network is a :class:`LogicNetwork` or
                       :class:`neet.boolean.WTNetwork`
        :param filename: filename to write drawing to. Temporary filename will
                         be used if no filename provided.
        :returns: a ``pygraphviz`` network drawing
        """
        nx.nx_agraph.view_pygraphviz(self.to_networkx_graph(
            labels=labels), prog='circo', path=filename)



"""
.. currentmodule:: neet.exceptions

Exceptions
==========

Exceptions are the key mechanism for handling undesirable program
state. Whenever :mod:`neet` encounters a problem, it raises an exception of
some variety. Whenever possible, we have preferred to use builtin exception
classes, e.g. `ValueError`, `IndexError`, etc... For causes that aren't really
covered by a builtin exception class, we've created subclasses of the standard
library's `Exception` to report those errors.

At the current point, there is only one such exception class,
:class:`FormatError`

API Documentation
-----------------
"""


class FormatError(Exception):
    """
    An error class to report when a configuration or data file is improperly
    formatted.
    """
    pass



import unittest
import neet.automata as ca
import numpy as np


class TestECA(unittest.TestCase):
    def test_is_network(self):
        from neet.interfaces import is_network
        self.assertTrue(is_network(ca.ECA))
        self.assertTrue(is_network(ca.ECA(23)))

    def test_is_not_fixed_sized(self):
        from neet.interfaces import is_fixed_sized
        self.assertFalse(is_fixed_sized(ca.ECA))
        self.assertFalse(is_fixed_sized(ca.ECA(23)))

    def test_fail_init(self):
        with self.assertRaises(ValueError):
            ca.ECA(-1)

        with self.assertRaises(ValueError):
            ca.ECA(256)

        with self.assertRaises(TypeError):
            ca.ECA([1, 1, 0, 1, 1, 0, 0, 1])

        with self.assertRaises(TypeError):
            ca.ECA("30")

        with self.assertRaises(TypeError):
            ca.ECA(30, boundary=[1, 2])

        with self.assertRaises(ValueError):
            ca.ECA(30, boundary=(1, 0, 1))

        with self.assertRaises(ValueError):
            ca.ECA(30, boundary=(1, 2))

    def test_init(self):
        for code in range(256):
            for left in range(2):
                for right in range(2):
                    eca = ca.ECA(code, (left, right))
                    self.assertEqual(code, eca.code)
                    self.assertEqual((left, right), eca.boundary)

    def test_invalid_code(self):
        eca = ca.ECA(30)

        eca.code = 45

        with self.assertRaises(ValueError):
            eca.code = -1

        with self.assertRaises(ValueError):
            eca.code = 256

        with self.assertRaises(TypeError):
            eca.code = "30"

    def test_invalid_boundary(self):
        eca = ca.ECA(30)

        eca.boundary = (0, 0)
        eca.boundary = None

        with self.assertRaises(ValueError):
            eca.boundary = (1, 1, 1)

        with self.assertRaises(ValueError):
            eca.boundary = (1, 2)

        with self.assertRaises(TypeError):
            eca.boundary = 1

        with self.assertRaises(TypeError):
            eca.boundary = [0, 1]

    def test_state_space(self):
        with self.assertRaises(ValueError):
            ca.ECA(30).state_space(0)

        with self.assertRaises(ValueError):
            ca.ECA(30).state_space(-1)

        eca = ca.ECA(30)
        self.assertEqual(2, len(list(eca.state_space(1))))
        self.assertEqual(4, len(list(eca.state_space(2))))
        self.assertEqual(8, len(list(eca.state_space(3))))

    def test_lattice_empty_update(self):
        eca = ca.ECA(30)
        with self.assertRaises(ValueError):
            eca.update([])

    def test_invalid_lattice_state_update(self):
        eca = ca.ECA(30)
        with self.assertRaises(ValueError):
            eca.update([-1, 0, 1])

        with self.assertRaises(ValueError):
            eca.update([1, 0, -1])

        with self.assertRaises(ValueError):
            eca.update([2, 0, 0])

        with self.assertRaises(ValueError):
            eca.update([1, 0, 2])

        with self.assertRaises(ValueError):
            eca.update([[1], [0], [2]])

        with self.assertRaises(ValueError):
            eca.update("101")

    def test_update_closed(self):
        eca = ca.ECA(30)

        lattice = [0]

        eca.update(lattice)
        self.assertEqual([0], lattice)

        lattice = [0, 0]

        eca.update(lattice)
        self.assertEqual([0, 0], lattice)

        lattice = [0, 0, 1, 0, 0]

        eca.update(lattice)
        self.assertEqual([0, 1, 1, 1, 0], lattice)

        eca.update(lattice)
        self.assertEqual([1, 1, 0, 0, 1], lattice)

    def test_update_open(self):
        eca = ca.ECA(30, (0, 1))

        lattice = [0]

        eca.update(lattice)
        self.assertEqual([1], lattice)

        lattice = [0, 0]

        eca.update(lattice)
        self.assertEqual([0, 1], lattice)

        lattice = [0, 0, 1, 0, 0]

        eca.update(lattice)
        self.assertEqual([0, 1, 1, 1, 1], lattice)

        eca.update(lattice)
        self.assertEqual([1, 1, 0, 0, 0], lattice)

    def test_update_long_time_closed(self):
        eca = ca.ECA(45)
        lattice = [1, 1, 0, 1, 0, 0, 1, 0, 1, 0, 0, 1, 0, 1]
        expected = [0, 1, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0]
        state_space = eca.state_space(len(lattice))
        if lattice in state_space:
            for n in range(1000):
                eca._unsafe_update(lattice)
        self.assertEqual(expected, lattice)

    def test_update_long_time_open(self):
        eca = ca.ECA(45, (0, 1))
        lattice = [1, 1, 0, 1, 0, 0, 1, 0, 1, 0, 0, 1, 0, 1]
        expected = [1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 1]
        state_space = eca.state_space(len(lattice))
        if lattice in state_space:
            for n in range(1000):
                eca._unsafe_update(lattice)
        self.assertEqual(expected, lattice)

    def test_update_numpy(self):
        eca = ca.ECA(30, (0, 1))

        lattice = np.asarray([0])

        eca.update(lattice)
        self.assertTrue(np.array_equal([1], lattice))

        lattice = np.asarray([0, 0])

        eca.update(lattice)
        self.assertTrue(np.array_equal([0, 1], lattice))

        lattice = [0, 0, 1, 0, 0]

        eca.update(lattice)
        self.assertTrue(np.array_equal([0, 1, 1, 1, 1], lattice))

        eca.update(lattice)
        self.assertTrue(np.array_equal([1, 1, 0, 0, 0], lattice))

    def test_update_index_error(self):
        eca = ca.ECA(30)
        with self.assertRaises(IndexError):
            eca.update([0, 0], index=2)

        with self.assertRaises(IndexError):
            eca.update([0, 0], index=-3)

    def test_update_index(self):
        eca = ca.ECA(30, (1, 1))

        lattice = [0, 0, 0, 0, 0]
        eca.update(lattice, index=0)
        self.assertEqual([1, 0, 0, 0, 0], lattice)

        lattice = [0, 0, 0, 0, 0]
        eca.update(lattice, index=1)
        self.assertEqual([0, 0, 0, 0, 0], lattice)

        lattice = [0, 0, 0, 0, 0]
        eca.update(lattice, index=-1)
        self.assertEqual([0, 0, 0, 0, 1], lattice)

        lattice = [0, 0, 1, 0, 0]
        eca.update(lattice, index=1)
        self.assertEqual([0, 1, 1, 0, 0], lattice)

    def test_update_index_numpy(self):
        eca = ca.ECA(30, (1, 1))

        lattice = np.asarray([0, 0, 0, 0, 0])
        eca.update(lattice, index=0)
        self.assertTrue(np.array_equal([1, 0, 0, 0, 0], lattice))

        lattice = np.asarray([0, 0, 0, 0, 0])
        eca.update(lattice, index=1)
        self.assertTrue(np.array_equal([0, 0, 0, 0, 0], lattice))

        lattice = np.asarray([0, 0, 0, 0, 0])
        eca.update(lattice, index=-1)
        self.assertTrue(np.array_equal([0, 0, 0, 0, 1], lattice))

        lattice = np.asarray([0, 0, 1, 0, 0])
        eca.update(lattice, index=1)
        self.assertTrue(np.array_equal([0, 1, 1, 0, 0], lattice))

    def test_update_pin_none(self):
        eca = ca.ECA(30)

        xs = [0, 0, 1, 0, 0]
        self.assertEqual([0, 1, 1, 1, 0], eca.update(xs, pin=None))
        self.assertEqual([1, 1, 0, 0, 1], eca.update(xs, pin=[]))

    def test_update_pin_index_clash(self):
        eca = ca.ECA(30)
        with self.assertRaises(ValueError):
            eca.update([0, 0], index=0, pin=[1])
        with self.assertRaises(ValueError):
            eca.update([0, 0], index=1, pin=[1])
        with self.assertRaises(ValueError):
            eca.update([0, 0], index=1, pin=[0, 1])

    def test_update_pin(self):
        eca = ca.ECA(30)

        xs = [0, 0, 1, 0, 0]
        self.assertEqual([0, 0, 1, 1, 0], eca.update(xs, pin=[1]))
        self.assertEqual([0, 0, 1, 0, 1], eca.update(xs, pin=[1]))
        self.assertEqual([1, 0, 1, 0, 1], eca.update(xs, pin=[1]))

        eca.boundary = (1, 1)
        xs = [0, 0, 0, 0, 0]
        self.assertEqual([1, 0, 0, 0, 0], eca.update(xs, pin=[-1]))
        self.assertEqual([1, 1, 0, 0, 0], eca.update(xs, pin=[0, -1]))

    def test_update_values_none(self):
        eca = ca.ECA(30)

        xs = [0, 0, 1, 0, 0]
        self.assertEqual([0, 1, 1, 1, 0], eca.update(xs, values=None))
        self.assertEqual([1, 1, 0, 0, 1], eca.update(xs, values={}))

    def test_update_invalid_values(self):
        eca = ca.ECA(30)
        with self.assertRaises(ValueError):
            eca.update([0, 0, 0, 0, 0], values={0: 2})
        with self.assertRaises(ValueError):
            eca.update([0, 0, 0, 0, 0], values={0: -1})

    def test_update_values_index_clash(self):
        eca = ca.ECA(30)
        with self.assertRaises(ValueError):
            eca.update([0, 0, 0, 0, 0], index=0, values={0: 1})
        with self.assertRaises(ValueError):
            eca.update([0, 0, 0, 0, 0], index=1, values={1: 0})
        with self.assertRaises(ValueError):
            eca.update([0, 0, 0, 0, 0], index=1, values={0: 0, 1: 0})

    def test_update_values_pin_clash(self):
        eca = ca.ECA(30)
        with self.assertRaises(ValueError):
            eca.update([0, 0, 0, 0, 0], pin=[0], values={0: 1})
        with self.assertRaises(ValueError):
            eca.update([0, 0, 0, 0, 0], pin=[1], values={1: 0})
        with self.assertRaises(ValueError):
            eca.update([0, 0, 0, 0, 0], pin=[1], values={0: 0, 1: 0})
        with self.assertRaises(ValueError):
            eca.update([0, 0, 0, 0, 0], pin=[1, 0], values={0: 0})

    def test_update_values(self):
        eca = ca.ECA(30)

        xs = [0, 0, 1, 0, 0]
        self.assertEqual([0, 1, 0, 1, 0], eca.update(xs, values={2: 0}))
        self.assertEqual([1, 0, 0, 0, 1], eca.update(xs, values={1: 0, 3: 0}))
        self.assertEqual([0, 1, 0, 1, 0], eca.update(xs, values={-1: 0}))

        eca.boundary = (1, 1)
        xs = [0, 0, 0, 0, 0]
        self.assertEqual([1, 0, 1, 0, 1], eca.update(xs, values={2: 1}))

    def test_neighbors_in(self):

        net = ca.ECA(30)

        self.assertEqual(net.neighbors_in(2, size=3), set([0, 1, 2]))

        with self.assertRaises(ValueError):
            self.assertEqual(net.neighbors_in(3, 3))

        net.boundary = (1, 1)

        self.assertEqual(net.neighbors_in(2, 3), set([1, 2, 3]))

        with self.assertRaises(ValueError):
            self.assertEqual(net.neighbors_in(3, 3))

        with self.assertRaises(ValueError):
            self.assertEqual(net.neighbors_in(5, 3))

        with self.assertRaises(TypeError):
            self.assertEqual(net.neighbors_in('2', 3))

        with self.assertRaises(ValueError):
            self.assertEqual(net.neighbors_in(-1, 3))

        with self.assertRaises(ValueError):
            self.assertEqual(net.neighbors_in(1, 0))

    def test_neighbors_out(self):

        net = ca.ECA(30)

        self.assertEqual(net.neighbors_out(2, 3), set([0, 1, 2]))

        with self.assertRaises(ValueError):
            self.assertEqual(net.neighbors_out(3, 3))

        net.boundary = (1, 1)

        self.assertEqual(net.neighbors_out(2, 3), set([1, 2]))

        with self.assertRaises(ValueError):
            self.assertEqual(net.neighbors_out(index=5, size=3))

        with self.assertRaises(TypeError):
            self.assertEqual(net.neighbors_out(index='2', size=3))

        with self.assertRaises(ValueError):
            self.assertEqual(net.neighbors_out(size=3, index=-1))

        with self.assertRaises(ValueError):
            self.assertEqual(net.neighbors_out(size=0, index=1))

    def test_neighbors_both(self):

        net = ca.ECA(30)

        self.assertEqual(net.neighbors(2, 4), set([1, 2, 3]))

    def test_to_networkx_metadata(self):
        net = ca.ECA(30)
        net.boundary = (1, 0)

        nx_net = net.to_networkx_graph(3)

        self.assertEqual(nx_net.graph['code'], 30)
        self.assertEqual(nx_net.graph['size'], 3)
        self.assertEqual(nx_net.graph['boundary'], (1, 0))



"""
.. currentmodule:: neet.boolean.examples

Example Networks
================

The examples module provides a collection of pre-loaded model networks such as
:attr:`s_pombe` (fission yeast) and :attr:`s_cerevisiae` (budding yeast).
"""
from neet.boolean import WTNetwork, LogicNetwork
from os.path import dirname, abspath, realpath, join

DATA_PATH = join(dirname(abspath(realpath(__file__))), "data")

S_POMBE_NODES = join(DATA_PATH, "s_pombe-nodes.txt")
S_POMBE_EDGES = join(DATA_PATH, "s_pombe-edges.txt")

#: The cell cycle network for *S. pombe* (fission yeast).
s_pombe = WTNetwork.read(S_POMBE_NODES, S_POMBE_EDGES)
s_pombe.metadata.update({
    'name': 's_pombe',
    'description': 'The cell cycle network for *S. pombe* (fission yeast).',
    'citation': '',
})

S_CEREVISIAE_NODES = join(DATA_PATH, "s_cerevisiae-nodes.txt")
S_CEREVISIAE_EDGES = join(DATA_PATH, "s_cerevisiae-edges.txt")

#: The cell cycle network for *S. cerevisiae* (budding yeast).
s_cerevisiae = WTNetwork.read(S_CEREVISIAE_NODES, S_CEREVISIAE_EDGES)
s_cerevisiae.metadata.update({
    'name': 's_cerevisiae',
    'description': 'The cell cycle network for *S. cerevisiae*'
                   ' (budding yeast).',
    'citation': '',
})

C_ELEGANS_NODES = join(DATA_PATH, "c_elegans-nodes.dat")
C_ELEGANS_EDGES = join(DATA_PATH, "c_elegans-edges.dat")

#: The cell cycle network for *C. elegans*.
c_elegans = WTNetwork.read(C_ELEGANS_NODES, C_ELEGANS_EDGES)
c_elegans.metadata.update({
    'name': 'c_elegans',
    'description': 'The cell cycle network for *C. elegans*.',
    'citation': '',
})

P53_NO_DMG_NODES = join(DATA_PATH, "p53_no_dmg-nodes.txt")
P53_NO_DMG_EDGES = join(DATA_PATH, "p53_no_dmg-edges.txt")

#: The p53 GRN with no damage present.
p53_no_dmg = WTNetwork.read(P53_NO_DMG_NODES, P53_NO_DMG_EDGES)
p53_no_dmg.metadata.update({
    'name': 'p53_no_dmg',
    'description': 'The p53 GRN with no damage present.',
    'citation': 'Choi, Minsoo, Jue Shi, Sung Hoon Jung, Xi Chen, and'
                ' Kwang-Hyun Cho. Attractor Landscape Analysis Reveals'
                ' Feedback Loops in the p53 Network That Control the'
                ' Cellular Response to DNA Damage. Science Signaling 5,'
                ' no. 251 (2012): ra83. doi:10.1126/scisignal.2003363.',
})

P53_DMG_NODES = join(DATA_PATH, "p53_dmg-nodes.txt")
P53_DMG_EDGES = join(DATA_PATH, "p53_dmg-edges.txt")

#: The p53 GRN with damage present.
p53_dmg = WTNetwork.read(P53_DMG_NODES, P53_DMG_EDGES)
p53_dmg.metadata.update({
    'name': 'p53_dmg',
    'description': 'The p53 GRN with damage present.',
    'citation': 'Choi, Minsoo, Jue Shi, Sung Hoon Jung, Xi Chen, and'
                ' Kwang-Hyun Cho. Attractor Landscape Analysis Reveals'
                ' Feedback Loops in the p53 Network That Control the'
                ' Cellular Response to DNA Damage. Science Signaling 5,'
                ' no. 251 (2012): ra83. doi:10.1126/scisignal.2003363.',
})

MOUSE_CORTICAL_7B_TRUTH_TABLE = join(
    DATA_PATH, "mouse_cortical_fig_7B-truth_table.txt")
MOUSE_CORTICAL_7B_EXPRESSIONS = join(
    DATA_PATH, "mouse_cortical_fig_7B-logic_expressions.txt")

#: The gene regulatory network for *mouse cortical* (7B edges)
mouse_cortical_7B = LogicNetwork.read_table(MOUSE_CORTICAL_7B_TRUTH_TABLE)
mouse_cortical_7B.metadata.update({
    'name': 'mouse_cortical_fig_7B',
    'description': 'The gene regulatory network for *mouse cortical*'
                   ' (Taken from figure 7B in citation below).',
    'citation': 'Giacomantonio, Clare E., and Geoffrey J. Goodhill. "A'
                ' Boolean model of the gene regulatory network underlying'
                ' Mammalian cortical area development." PLoS'
                ' computational biology 6, no. 9 (2010):'
                ' e1000936. doi:10.1371/journal.pcbi.1000936',

})
mouse_cortical_7B_from_expr = LogicNetwork.read_logic(
    MOUSE_CORTICAL_7B_EXPRESSIONS)

MOUSE_CORTICAL_7C_TRUTH_TABLE = join(
    DATA_PATH, "mouse_cortical_fig_7C-truth_table.txt")
MOUSE_CORTICAL_7C_EXPRESSIONS = join(
    DATA_PATH, "mouse_cortical_fig_7C-logic_expressions.txt")

#: The gene regulatory network for *mouse cortical* (7C edges)
mouse_cortical_7C = LogicNetwork.read_table(MOUSE_CORTICAL_7C_TRUTH_TABLE)
mouse_cortical_7C.metadata.update({
    'name': 'mouse_cortical_fig_7C',
    'description': 'The gene regulatory network for *mouse cortical*'
                   ' (Taken from figure 7C in citation below).',
    'citation': 'Giacomantonio, Clare E., and Geoffrey J. Goodhill. "A'
                ' Boolean model of the gene regulatory network underlying'
                ' Mammalian cortical area development." PLoS'
                ' computational biology 6, no. 9 (2010):'
                ' e1000936. doi:10.1371/journal.pcbi.1000936',

})
mouse_cortical_7C_from_expr = LogicNetwork.read_logic(
    MOUSE_CORTICAL_7C_EXPRESSIONS)

MYELOID_TRUTH_TABLE = join(DATA_PATH, "myeloid-truth_table.txt")
MYELOID_LOGIC_EXPRESSIONS = join(DATA_PATH, "myeloid-logic_expressions.txt")

#: Differentiation control network for *myeloid* progenitors.
myeloid = LogicNetwork.read_table(MYELOID_TRUTH_TABLE)
myeloid.metadata.update({
    'name': 'myeloid',
    'description': 'Differentiation control network for *myeloid*'
                   ' progenitors.',
    'citation': '',
})
myeloid_from_expr = LogicNetwork.read_logic(MYELOID_LOGIC_EXPRESSIONS)



"""
.. currentmodule:: neet.interfaces

.. testsetup:: interfaces

    from neet.automata import ECA
    from neet.interfaces import *
    from neet.statespace import StateSpace

Interfaces
==========

The :mod:`neet.interfaces` module provides a collection of functions for
determining if types adhere to various network interfaces, and generic
functions for operating upon them. This done primarily through the
:func:`is_network`, :func:`is_fixed_sized` and :func:`is_boolean_network`
functions.

API Documentation
-----------------
"""


def is_network(thing):
    """
    Determine whether an *object* or *type* meets the interface requirement of
    a network. Specifically, to be considered a network, a class must provide
    the following methods:

    1. `update` which updates the state of a lattice
    2. `state_space` which returns a :func:`neet.statespace.StateSpace` object
    3. `neighbors` which returns the neighbors of a given node

    .. rubric:: Example:

    .. doctest:: interfaces

        >>> class IsNetwork(object):
        ...     def update(self):
        ...         pass
        ...     def state_space(self):
        ...         return StateSpace(1)
        ...     def neighbors(self, i):
        ...         return []
        ...
        >>> class IsNotNetwork(object):
        ...     pass
        ...
        >>> is_network(IsNetwork())
        True
        >>> is_network(IsNetwork)
        True
        >>> is_network(IsNotNetwork())
        False
        >>> is_network(IsNotNetwork)
        False
        >>> is_network(5)
        False

    :param thing: an object or a type
    :returns: ``True`` if ``thing`` has the minimum interface of a network
    """
    has_update = hasattr(thing, 'update')
    has_state_space = hasattr(thing, 'state_space')
    has_neighbors = hasattr(thing, 'neighbors')

    return has_update and has_state_space and has_neighbors


def is_fixed_sized(thing):
    """
    Determine whether an *object* or *type* is a network and has a fixed size.

    .. rubric:: Example

    .. doctest:: interfaces

        >>> class IsNetwork(object):
        ...     def update(self):
        ...         pass
        ...     def state_space(self):
        ...         return StateSpace(1)
        ...     def neighbors(self, i):
        ...         return []
        ...
        >>> class FixedSized(IsNetwork):
        ...     def size():
        ...         return 5
        ...
        >>> is_fixed_sized(IsNetwork)
        False
        >>> is_fixed_sized(FixedSized)
        True

    :param thing: an object or a type
    :returns: ``True`` if ``thing`` is a network with a size attribute
    :see: :func:`is_network`.
    """
    return is_network(thing) and hasattr(thing, 'size')


def is_boolean_network(thing):
    """
    Determine whether an *object* is a network with all Boolean states.d
    """
    # Boolean networks have a single base equal to 2
    if is_network(thing) and hasattr(thing.state_space(), 'base'):
        return thing.state_space().base == 2
    else:
        return False


def neighbors(net, index, direction='both', **kwargs):
    """
    Return a set of neighbors for a specified node, or a list of sets of
    neighbors for all nodes in the network.

    For ECAs, in the cases of the lattices having fixed boundary conditions,
    the left boundary, being on the left of the leftmost index 0, has an index
    of -1, while the right boundary's index is the size+1. The full state of
    the lattices and the boundaries is equavolent to: `[cell0, cell1, ...,
    cellN, right_boundary, left_boundary]` if it is ever presented as a single
    list in Python.

    :param index: node index, if neighbors desired for one node only
    :param direction: type of node neighbors to return ('in', 'out', or 'both')
    :kwarg size: size of ECA, required if network is an ECA
    :returns: a set of neighbors of a node

    .. rubric:: Example

    .. doctest:: interfaces

        >>> net = ECA(30)
        >>> neighbors(net, index=2, size=3, direction='out')
        {0, 1, 2}
        >>> net.boundary = (1,1)
        >>> neighbors(net, index=2, size=3, direction='out')
        {1, 2}

    See `ECA.neighbors()`,`LogicNetwork.neighbors()` or `WTNetwork.neighbors()`
    docstrings for more details and basic use examples.

    """
    if direction not in ('in', 'out', 'both'):
        raise ValueError('direction must be "in", "out" or "both"')

    neighbor_types = {'in': net.neighbors_in,
                      'out': net.neighbors_out,
                      'both': net.neighbors}

    if net.__class__.__name__ == 'ECA':
        if 'size' not in kwargs:
            msg = "A `size` kwarg is required for returning an ECA's neighbors"
            raise AttributeError(msg)
        else:
            return neighbor_types[direction](index, size=kwargs['size'])

    else:
        return neighbor_types[direction](index)


def to_networkx_graph(net, size=None, labels='indices', **kwargs):
    """
    Return networkx graph given neet network. Requires `networkx`.

    :param labels: how node is labeled and thus identified in networkx graph
                   ('names' or 'indices'), only used if `net` is a
                   `LogicNetwork` or `WTNetwork`
    :kwarg size: size of ECA, required if network is an ECA
    :returns: a networkx DiGraph
    """
    if net.__class__.__name__ == 'ECA':
        if size is None:
            msg = "`size` required to convert an ECA to a networkx network"
            raise AttributeError(msg)
        else:
            return net.to_networkx_graph(size)

    elif net.__class__.__name__ in ['WTNetwork', 'LogicNetwork']:
        return net.to_networkx_graph(labels=labels)



from setuptools import setup

with open("README.md") as f:
    README = f.read()

with open("LICENSE") as f:
    LICENSE = f.read()

setup(
    name='neet',
    version='0.1.0',
    description='A brilliant and fundamental contribution to network science',
    long_description=README,
    maintainer='Douglas G. Moore',
    maintainer_email='doug@dglmoore.com',
    url='https://github.com/elife-asu/neet',
    license=LICENSE,
    install_requires=['numpy', 'networkx', 'pyinform'],
    packages=['neet', 'neet.automata', 'neet.boolean'],
    package_data={'neet.boolean': ['data/*.txt', 'data/*.dat']},
    test_suite='test',
    platforms=['Windows', 'OS X', 'Linux']
)



import unittest
from neet.python import long
from neet.statespace import StateSpace
import numpy as np


class TestStateSpace(unittest.TestCase):
    def test_invalid_spec_type(self):
        with self.assertRaises(TypeError):
            StateSpace("a")

        with self.assertRaises(TypeError):
            StateSpace("abc")

    def test_invalid_base_type(self):
        with self.assertRaises(TypeError):
            StateSpace(5, base='a')

        with self.assertRaises(TypeError):
            StateSpace(3, base=2.5)

        with self.assertRaises(TypeError):
            StateSpace([1.0, 2.0, 3.0])

    def test_invalid_spec_value(self):
        with self.assertRaises(ValueError):
            StateSpace(0)

        with self.assertRaises(ValueError):
            StateSpace(-1)

        with self.assertRaises(ValueError):
            StateSpace([])

        with self.assertRaises(ValueError):
            StateSpace([0])

        with self.assertRaises(ValueError):
            StateSpace([-1])

    def test_invalid_base_value(self):
        with self.assertRaises(ValueError):
            StateSpace(3, base=0)

        with self.assertRaises(ValueError):
            StateSpace(4, base=-1)

    def test_uniform_bases(self):
        spec = StateSpace(5)
        self.assertTrue(spec.is_uniform)
        self.assertEqual(5, spec.ndim)
        self.assertEqual(2, spec.base)
        self.assertEqual(32, spec.volume)

        spec = StateSpace(8, base=4)
        self.assertTrue(spec.is_uniform)
        self.assertEqual(8, spec.ndim)
        self.assertEqual(4, spec.base)
        self.assertEqual(65536, spec.volume)

        spec = StateSpace([3, 3, 3, 3])
        self.assertTrue(spec.is_uniform)
        self.assertEqual(4, spec.ndim)
        self.assertEqual(3, spec.base)
        self.assertEqual(81, spec.volume)

    def test_base_mismatch(self):
        with self.assertRaises(ValueError):
            StateSpace([2, 2, 2], base=3)

        with self.assertRaises(ValueError):
            StateSpace([3, 3, 3], base=2)

        with self.assertRaises(ValueError):
            StateSpace([2, 2, 3], base=2)

        with self.assertRaises(ValueError):
            StateSpace([2, 2, 3], base=3)

        StateSpace([2, 2, 2], base=2)
        StateSpace([3, 3, 3], base=3)

    def test_nonuniform_bases(self):
        spec = StateSpace([1, 2, 3, 2, 1])
        self.assertFalse(spec.is_uniform)
        self.assertEqual([1, 2, 3, 2, 1], spec.base)
        self.assertEqual(5, spec.ndim)
        self.assertEqual(12, spec.volume)

    def test_states_boolean(self):
        space = StateSpace(1)
        self.assertEqual([[0], [1]],
                         list(space))

        space = StateSpace(2)
        self.assertEqual([[0, 0], [1, 0], [0, 1], [1, 1]],
                         list(space))

        space = StateSpace(3)
        self.assertEqual([[0, 0, 0], [1, 0, 0], [0, 1, 0], [1, 1, 0],
                          [0, 0, 1], [1, 0, 1], [0, 1, 1], [1, 1, 1]],
                         list(space))

    def test_states_nonboolean(self):
        space = StateSpace(1, base=1)
        self.assertEqual([[0]],
                         list(space))

        space = StateSpace(1, base=3)
        self.assertEqual([[0], [1], [2]],
                         list(space))

        space = StateSpace(2, base=1)
        self.assertEqual([[0, 0]],
                         list(space))

        space = StateSpace(2, base=3)
        self.assertEqual([[0, 0], [1, 0], [2, 0],
                          [0, 1], [1, 1], [2, 1],
                          [0, 2], [1, 2], [2, 2]],
                         list(space))

    def test_states_boolean_list(self):
        space = StateSpace([2])
        self.assertEqual([[0], [1]],
                         list(space))

        space = StateSpace([2, 2])
        self.assertEqual([[0, 0], [1, 0], [0, 1], [1, 1]],
                         list(space))

        space = StateSpace([2, 2, 2])
        self.assertEqual([[0, 0, 0], [1, 0, 0], [0, 1, 0], [1, 1, 0],
                          [0, 0, 1], [1, 0, 1], [0, 1, 1], [1, 1, 1]],
                         list(space))

    def test_states_nonboolean_list(self):
        space = StateSpace([1])
        self.assertEqual([[0]],
                         list(space))

        space = StateSpace([3])
        self.assertEqual([[0], [1], [2]],
                         list(space))

        space = StateSpace([1, 2])
        self.assertEqual([[0, 0], [0, 1]],
                         list(space))

        space = StateSpace([1, 3])
        self.assertEqual([[0, 0], [0, 1], [0, 2]],
                         list(space))

        space = StateSpace([3, 3])
        self.assertEqual([[0, 0], [1, 0], [2, 0],
                          [0, 1], [1, 1], [2, 1],
                          [0, 2], [1, 2], [2, 2]],
                         list(space))

    def test_states_count(self):
        xs = [3, 5, 2, 5, 2, 1, 4, 2]
        space = StateSpace(xs)
        count = 0
        for state in space:
            count += 1
        self.assertEqual(np.product(xs), count)

    def test_encoding_error(self):
        space = StateSpace(3)
        with self.assertRaises(ValueError):
            space.encode([1, 1])

        space = StateSpace(1)
        with self.assertRaises(ValueError):
            space.encode([2])

        space = StateSpace([2, 3])
        with self.assertRaises(ValueError):
            space.encode([1, 3])

        with self.assertRaises(ValueError):
            space.encode([1, -1])

    def test_encoding_uniform(self):
        for width in range(1, 5):
            for base in range(1, 5):
                space = StateSpace(width, base)
                counter = 0
                for state in space:
                    encoding = space.encode(state)
                    self.assertEqual(counter, encoding)
                    counter += 1

    def test_encoding_nonuniform(self):
        for a in range(1, 5):
            for b in range(1, 5):
                for c in range(1, 5):
                    space = StateSpace([a, b, c])
                    counter = 0
                    for state in space:
                        encoding = space.encode(state)
                        self.assertEqual(counter, encoding)
                        counter += 1

    def test_decoding_uniform(self):
        for width in range(1, 5):
            for base in range(1, 5):
                space = StateSpace(width, base)
                states = list(space)
                decoded = list(map(space.decode, range(space.volume)))
                self.assertEqual(states, decoded)

    def test_decoding_nonuniform(self):
        for a in range(1, 5):
            for b in range(1, 5):
                for c in range(1, 5):
                    space = StateSpace([a, b, c])
                    states = list(space)
                    decoded = list(map(space.decode, range(space.volume)))
                    self.assertEqual(states, decoded)

    def test_encode_decode_uniform(self):
        for width in range(1, 5):
            for base in range(1, 5):
                space = StateSpace(width, base)
                for state in space:
                    encoded = space.encode(state)
                    decoded = space.decode(encoded)
                    self.assertEqual(state, decoded)

    def test_encode_decode_nonuniform(self):
        for a in range(1, 5):
            for b in range(1, 5):
                for c in range(1, 5):
                    space = StateSpace([a, b, c])
                    for state in space:
                        encoded = space.encode(state)
                        decoded = space.decode(encoded)
                        self.assertEqual(state, decoded)

    def test_decode_encode_uniform(self):
        for width in range(1, 5):
            for base in range(1, 5):
                space = StateSpace(width, base)
                for i in range(base**width):
                    decoded = space.decode(i)
                    encoded = space.encode(decoded)
                    self.assertEqual(i, encoded)

    def test_decode_encode_nonuniform(self):
        for a in range(1, 5):
            for b in range(1, 5):
                for c in range(1, 5):
                    space = StateSpace([a, b, c])
                    for i in range(a * b * c):
                        decoded = space.decode(i)
                        encoded = space.encode(decoded)
                        self.assertEqual(i, encoded)

    def test_check_states_uniform(self):
        state_space = StateSpace(3)
        self.assertTrue([0, 1, 1] in state_space)
        self.assertFalse([0, 0] in state_space)
        self.assertFalse([1, 2, 0] in state_space)

        self.assertFalse([0, 1, 1] not in state_space)
        self.assertTrue([0, 0] not in state_space)
        self.assertTrue([1, 2, 0] not in state_space)

    def test_check_states_varied(self):
        self.assertTrue([0, 2, 1] in StateSpace([2, 3, 2]))
        self.assertFalse([0, 1] in StateSpace([2, 2, 3]))
        self.assertFalse([1, 1, 6] in StateSpace([2, 3, 4]))

        self.assertFalse([0, 2, 1] not in StateSpace([2, 3, 2]))
        self.assertTrue([0, 1] not in StateSpace([2, 2, 3]))
        self.assertTrue([1, 1, 6] not in StateSpace([2, 3, 4]))

    def test_long_encoding(self):
        state_space = StateSpace(10)
        code = state_space.encode(np.ones(10, dtype=int))
        print(type(code))
        self.assertIsInstance(code, long)

        state_space = StateSpace(68)
        code = state_space.encode(np.ones(68, dtype=int))
        self.assertIsInstance(code, long)

        state_space = StateSpace(100)
        code = state_space.encode(np.ones(100, dtype=int))
        self.assertIsInstance(code, long)



import unittest
from neet.sensitivity import (sensitivity, average_sensitivity, lambdaQ,
                              is_canalizing, canalizing_edges,
                              canalizing_nodes)
from neet.sensitivity import _hamming_neighbors
from neet.statespace import StateSpace
import neet.boolean as NB
import numpy as np


class TestSensitivityWTNetwork(unittest.TestCase):
    class IsBooleanNetwork(object):
        def update(self, lattice):
            pass

        def state_space(self):
            return StateSpace(1)

    class IsNotBooleanNetwork(object):
        def update(self, lattice):
            pass

        def state_space(self):
            return StateSpace([2, 3, 4])

    class IsNotNetwork(object):
        pass

    def test_sensitivity_net_type(self):
        with self.assertRaises(TypeError):
            sensitivity(self.IsNotNetwork(), [0, 0, 0])

        with self.assertRaises(TypeError):
            sensitivity(self.IsNotBooleanNetwork(), [0, 0, 0])

    def test_hamming_neighbors_input(self):
        with self.assertRaises(ValueError):
            _hamming_neighbors([0, 1, 2])

        with self.assertRaises(ValueError):
            _hamming_neighbors([[0, 0, 1], [1, 0, 0]])

    def test_hamming_neighbors_example(self):
        state = [0, 1, 1, 0]
        neighbors = [[1, 1, 1, 0],
                     [0, 0, 1, 0],
                     [0, 1, 0, 0],
                     [0, 1, 1, 1]]
        self.assertTrue(np.array_equal(neighbors, _hamming_neighbors(state)))

    def test_sensitivity(self):
        net = NB.WTNetwork([[1, -1], [0, 1]], [0.5, 0])
        self.assertEqual(1.0, sensitivity(net, [0, 0]))

    def test_average_sensitivity_net_type(self):
        with self.assertRaises(TypeError):
            average_sensitivity(self.IsNotNetwork())

        with self.assertRaises(TypeError):
            average_sensitivity(self.IsNotBooleanNetwork())

    def test_average_sensitivity_lengths(self):
        net = NB.WTNetwork([[1, -1], [0, 1]], [0.5, 0])

        with self.assertRaises(ValueError):
            average_sensitivity(
                net, states=[[0, 0], [0, 1]], weights=[0, 1, 2])

    def test_average_sensitivity(self):
        net = NB.WTNetwork([[1, -1], [0, 1]], [0.5, 0])
        self.assertEqual(1.0, average_sensitivity(net))

    def test_sensitivity_s_pombe(self):
        from neet.boolean.examples import s_pombe
        s = sensitivity(s_pombe, [0, 0, 0, 0, 0, 1, 1, 0, 0])
        self.assertAlmostEqual(s, 1.0)

    def test_average_sensitivity_c_elegans(self):
        from neet.boolean.examples import c_elegans

        s = average_sensitivity(c_elegans)
        self.assertAlmostEqual(s, 1.265625)

        s = average_sensitivity(c_elegans,
                                states=[[0, 0, 0, 0, 0, 0, 0, 0],
                                        [1, 1, 1, 1, 1, 1, 1, 1]],
                                weights=[9, 1])
        self.assertAlmostEqual(s, 1.7)

    def test_lambdaQ_c_elegans(self):
        from neet.boolean.examples import c_elegans
        self.assertAlmostEqual(lambdaQ(c_elegans), 1.263099227661824)

    def test_average_sensitivity_logic_network(self):
        net = NB.LogicNetwork([((1, 2), {'01', '10'}),
                               ((0, 2), ('01', '10', '11')),
                               ((0, 1), {'11'})])

        s = average_sensitivity(net)
        self.assertAlmostEqual(s, 1.3333333333333333)

        s = average_sensitivity(net, weights=np.ones(8))
        self.assertAlmostEqual(s, 1.3333333333333333)

        s = average_sensitivity(net, states=net.state_space())
        self.assertAlmostEqual(s, 1.3333333333333333)

    def test_lambdaQ_logic_network(self):
        net = NB.LogicNetwork([((1, 2), {'01', '10'}),
                               ((0, 2), ('01', '10', '11')),
                               ((0, 1), {'11'})])
        self.assertAlmostEqual(lambdaQ(net), 1.2807764064044149)

    def test_is_canalizing_logic_network(self):
        net = NB.LogicNetwork([((1, 2), {'01', '10'}),
                               ((0, 2), ('01', '10', '11')),
                               ((0, 1), {'11'})])

        self.assertFalse(is_canalizing(net, 0, 1))
        self.assertTrue(is_canalizing(net, 1, 0))
        self.assertTrue(is_canalizing(net, 2, 1))

    def test_canalizing(self):
        net = NB.LogicNetwork([((1, 2), {'01', '10'}),
                               ((0, 2), ('01', '10', '11')),
                               ((0, 1), {'11'})])

        edges = canalizing_edges(net)
        self.assertEqual(edges, {(1, 0), (1, 2), (2, 0), (2, 1)})

        nodes = canalizing_nodes(net)
        self.assertEqual(nodes, {1, 2})



import unittest
import networkx as nx
import numpy as np
from neet.boolean.examples import mouse_cortical_7B
from neet.boolean.examples import s_pombe
from neet.boolean.randomnet import (random_logic, random_binary_states,)

TESTSEED = 314159

class TestRandomnet(unittest.TestCase):

    def test_random_logic_invalid_p(self):
        """
        ``random_logic`` should raise a value error if ``p`` is an
        incorrect size
        """
        with self.assertRaises(ValueError):
            net = mouse_cortical_7B
            random_logic(net, p=np.ones(net.size + 1))

    def test_random_binary_states(self):
        self.assertEqual(8, len(random_binary_states(4, 0.5)))
        self.assertTrue(len(random_binary_states(3, 0.4)) in (3, 4))

    def test_random_logic_fixed_structure(self):
        net = mouse_cortical_7B
        np.random.seed(TESTSEED)
        randnet = random_logic(net, connections='fixed-structure')
        # fixed-structure should preserve all neighbors
        for i in range(net.size):
            self.assertEqual(net.neighbors_in(i), randnet.neighbors_in(i))

    def test_random_logic_fixed_in_degree(self):
        net = mouse_cortical_7B
        np.random.seed(TESTSEED)
        randnet = random_logic(net, connections='fixed-in-degree')
        # fixed-in-degree should preserve each node's in degree
        for i in range(net.size):
            self.assertEqual(len(net.neighbors_in(i)),
                             len(randnet.neighbors_in(i)))

    def test_random_logic_fixed_mean_degree(self):
        net = mouse_cortical_7B
        np.random.seed(TESTSEED)
        randnet = random_logic(net, connections='fixed-mean-degree')
        # fixed-mean-degree should preserve the total number of edges
        numedges = np.sum([len(net.neighbors_in(i)) for i in range(net.size)])
        randnumedges = np.sum([len(randnet.neighbors_in(i))
                               for i in range(randnet.size)])
        self.assertEqual(numedges, randnumedges)
